using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;

namespace HQF
{
    public static class Extensions
    {
        public static void RemoveAllChildren(this Transform self)
        {
            var children = new List<GameObject>();
            foreach (Transform child in self.transform) {
                children.Add(child.gameObject);
            }
            // same:
            // children.ForEach(child => UnityEngine.Object.Destroy(child));
            children.ForEach(UnityEngine.Object.Destroy);
        }

        public static GameObject FindParentWithTag(
            this GameObject self, string tag
        ) {
            var parent = self.transform.parent;
            while (parent != null) { 
                if (parent.CompareTag(tag)) {
                    return parent.gameObject;
                }
                parent = parent.transform.parent;
            }
            return null;
        }
        
        /// <summary>
        ///   <para>Rt = RectTransform</para>
        /// </summary>
        public static RectTransform Rt(this GameObject self)
        {
            return self.GetComponent<RectTransform>();
        }

        /// <summary>
        ///   <para>FdC = find deep child (for GameObject object)</para>
        /// </summary>
        public static GameObject FdC(this GameObject self, string name)
        {
            return self.transform.FindDeepChild(name).gameObject;
        }

        /// <summary>
        ///   <para>FdC = find deep child (for Transform object)</para>
        /// </summary>
        public static GameObject FdC(this Transform self, string name)
        {
            return self.gameObject.transform.FindDeepChild(name).gameObject;
        }

        /// <summary>
        ///   <para>Gc = GetComponent&lt;>()</para>
        /// </summary>
        public static T Gc<T>(this GameObject self)
        {
            return self.GetComponent<T>();
        }

        /// <summary>
        ///   <para>Gc = GetComponent&lt;>()</para>
        /// </summary>
        public static T Gc<T>(this Transform self)
        {
            return self.GetComponent<T>();
        }

        /// <summary>
        ///   <para>GcC = get component in children</para>
        /// </summary>
        public static T GcC<T>(this GameObject self)
        {
            return self.GetComponentInChildren<T>();
        }

        /// <summary>
        ///   <para>GcCs = get components in children</para>
        /// </summary>
        public static T[] GcCs<T>(this GameObject self)
        {
            return self.GetComponentsInChildren<T>();
        }

        /// <summary>
        ///   <para>GcC = get component in parent</para>
        /// </summary>
        public static T GcP<T>(this GameObject self)
        {
            return self.GetComponentInParent<T>();
        }

        /// <summary>
        ///   <para>Txt = find txt</para>
        /// </summary>
        public static Text Txt(this GameObject self)
        {
            return self.GetComponent<Text>();
        }

        /// <summary>
        ///   <para>Img = find Image</para>
        /// </summary>
        public static Image Img(this GameObject self)
        {
            return self.GetComponent<Image>();
        }

        public static Transform FindDeepChild(this Transform self, string name)
        {
            // Breadth-first search
            var result = self.Find(name);
            if (result != null)
                return result;
            foreach (Transform child in self) {
                result = child.FindDeepChild(name);
                if (result != null)
                    return result;
            }
            return null;
            /*
                // Depth-first search
                foreach(Transform child in self) {
                    if (child.name == name ) {
                        return child;
                    }
                    var result = child.FindDeepChild(name);
                    if (result != null) {
                        return result;
                    }
                }
                return null;
            */
        }

        /// <summary>
        ///   <para>"Next()" on enum types</para>
        /// </summary>
        public static T Next<T>(this T src) where T : struct
        {
            /**
             * Example:
             * public enum BoatOrientation { A, B, C, D };
             * BoatOrientation test = BoatOrientation.A;
             * test = test.Next();
             * 
             */
            if (!typeof(T).IsEnum) {
                throw new ArgumentException(string.Format(
                    "Argument {0} is not an Enum", typeof(T).FullName
                ));
            }
            T[] arr = (T[])Enum.GetValues(src.GetType());
            int j = Array.IndexOf(arr, src);
            if (j < 0) {
                throw new IndexOutOfRangeException();
            }
            j++;
            return (arr.Length == j) ? arr[0] : arr[j];
        }

        /// <summary>
        ///   <para>"Prev()" on enum types</para>
        /// </summary>
        public static T Prev<T>(this T src) where T : struct
        {
            /**
             * Example:
             * public enum BoatOrientation { A, B, C, D };
             * BoatOrientation test = BoatOrientation.A;
             * test = test.Prev();
             * 
             */
            if (!typeof(T).IsEnum) {
                throw new ArgumentException(string.Format(
                    "Argument {0} is not an Enum", typeof(T).FullName
                ));
            }
            T[] arr = (T[])Enum.GetValues(src.GetType());
            int j = Array.IndexOf(arr, src);
            if (j<0) {
                throw new IndexOutOfRangeException();
            }
            j--;
            return (j < 0) ? arr[arr.Length-1] : arr[j];
        }
    }
}
